﻿/*
This file is part of My Nes
A Nintendo Entertainment System Emulator.

 Copyright © 2009 - 2010 Ala Hadid (AHD)

My Nes is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

My Nes is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using MyNes.Nes.Input;
using MyNes.Nes.Output.Video;
using MyNes.Nes.Output.Audio;
namespace MyNes.Nes
{
    /// <summary>
    /// The main class of nes machine we emulate
    /// </summary>
    public class NesEngine
    {
        Memory _MEMORY;
        CPU _CPU;
        PPU _PPU;
        APU _APU;
        TVFORMAT _TvFormat = TVFORMAT.NTSC;
        bool _SoundEnabled;
        bool _AutoSaveSRAM = true;
        public NesEngine(TVFORMAT TvFormat, PaletteFormat PlFormat)
        {
            DEBUG.WriteLine(this, "Initializeing the nes emulation engine...", DebugStatus.None);
            _TvFormat = TvFormat;
            //Initialize Engine
            _MEMORY = new Memory(this);
            _CPU = new CPU(_MEMORY, TvFormat, this);
            _CPU.PauseToggle += new EventHandler<EventArgs>(_CPU_PauseToggle);
            _PPU = new PPU(TvFormat, PlFormat, this);
            DEBUG.WriteLine(this, "Nes initialized ok.", DebugStatus.Cool);
        }
        void _CPU_PauseToggle(object sender, EventArgs e)
        {
            //Rise the pause event
            if (PauseToggle != null)
                PauseToggle(this, null);
        }
        /// <summary>
        /// Load a rom into the memory
        /// </summary>
        /// <param name="RomPath">The INES rom path</param>
        public bool LoadRom(string RomPath)
        {
            return _MEMORY.LoadCart(RomPath);
        }
        /// <summary>
        /// Run the Nes
        /// </summary>
        public void Run()
        {
            _CPU.Reset();
            _CPU.ON = true;
            _CPU.Run();
            _CPU.ON = false;
        }
        /// <summary>
        /// Toggle pause the nes
        /// </summary>
        public void TogglePause()
        {
            _CPU.TogglePause();
        }
        public void Pause()
        {
            _CPU.Pause = true;
            //Rise the pause event
            if (PauseToggle != null)
                PauseToggle(this, null);
        }
        public void Resume()
        {
            _CPU.Pause = false;
            //Rise the pause event
            if (PauseToggle != null)
                PauseToggle(this, null);
        }
        public void ShutDown()
        {
            while (_APU.IsRendering)
            { }
            _APU.Pause();
            _APU.Shutdown();
            if (_APU.RECODER.IsRecording)
                _APU.RECODER.Stop();
            if (_MEMORY.MAP.Cartridge.IsSaveRam & _AutoSaveSRAM)
                SaveSRAM(_MEMORY.MAP.Cartridge.saveFilename);
            _CPU.ON = false;
        }
        public void SoftReset()
        {
            _CPU.SoftReset();
        }
        public void SaveSRAM(string FilePath)
        {
            //If we have save RAM, try to save it
            try
            {
                using (FileStream writer = File.OpenWrite(FilePath))
                {
                    writer.Write(_MEMORY.SRAM, 0, 0x2000);
                    DEBUG.WriteLine(this, "SRAM saved !!", DebugStatus.Cool);
                }
            }
            catch
            {
                DEBUG.WriteLine(this, "Could not save S-RAM.", DebugStatus.Error);
            }
        }
        public void SetupInput(InputManager Manager, Joypad Joy1, Joypad Joy2)
        {
            _MEMORY.InputManager = Manager;
            _MEMORY.Joypad1 = Joy1;
            _MEMORY.Joypad2 = Joy2;
        }
        public void SetupOutput(IGraphicDevice VideoDevice, IAudioDevice AudioDevice)
        {
            _PPU.OutputDevice = VideoDevice;
            _APU = new APU(this, AudioDevice);
        }
        public void SetupTV(TVFORMAT TvFormat, PaletteFormat PLFormat)
        {
            Pause();
            _CPU.SetTVFormat(TvFormat);
            _PPU.SetTVFormat(TvFormat, PLFormat);
        }
        //Properties
        public TVFORMAT TvFormat
        { get { return _TvFormat; } }
        public CPU CPU
        { get { return _CPU; } }
        public Memory MEMORY
        { get { return _MEMORY; } }
        public PPU PPU
        { get { return _PPU; } set { _PPU = value; } }
        public APU APU
        { get { return _APU; } set { _APU = value; } }
        public bool SoundEnabled
        { get { return _SoundEnabled; } set { _SoundEnabled = value; } }
        public event EventHandler<EventArgs> PauseToggle;
        public bool AutoSaveSRAM
        { get { return _AutoSaveSRAM; } set { _AutoSaveSRAM = value; } }
    }
    public enum TVFORMAT
    { NTSC, PAL }
}
